r"""
.. currentmodule:: galgebra

========================================
galgebra (:mod:`galgebra`)
========================================

A module for geometric algebra.

Submodules
==========

.. autosummary::
    :toctree: generated

    metric
    ga
    mv
    lt
    dop
    atoms
    printer
    utils

.. note::

    Some functions may be better documented by the original non-generated docs
    on the :doc:`module-components` page, but there is a risk the descriptions
    there are outdated.

    If you would like to help with merging the descriptions here with a
    description on that page, please head over to :issue:`300` on GitHub, where
    there's an explanation of how to do so. Even merging just one function
    explanation helps!
"""

from ._version import __version__  # noqa: F401
